{-----------------------------------------------------------------------------

	         A LIBRARY OF FUNCTIONAL STRATEGY COMBINATORS

		       		  StrategyLib

                   Ralf Laemmel                Joost Visser
               CWI & VU, Amsterdam         CWI & SIG, Amsterdam

This module is part of a library of functional strategy combinators,
including combinators for generic traversal. This module provides
combinators to localize monadic effects.

-----------------------------------------------------------------------------} 

module EffectTheme (
	module EffectTheme
) where

import Monad
import MonadRun
import MonadTrans
import StrategyPrelude
import OverloadingTheme
import StrategyPrimitives


--- Replace one strategy monad by another ------------------------------------

-- The two monads are unrelated.
-- The first is run, and the resulting value is returned in the other.
-- We use the `run' function of one monad, and the `return' function
-- of the other.

mrunTP 		:: (Monad m, Monad m', MonadRun s m) 
		=> (forall a . s a a) -> TP m -> TP m'
mrunTP alg	=  msubstTP (mrun alg)

mrunTU 		:: (Monad m, Monad m', MonadRun s m) 
		=> s a a -> TU a m -> TU a m'
mrunTU alg	=  msubstTU (mrun alg)


--- Add an effect to the strategy monad --------------------------------------

-- The second monad is obtained from the first 
-- via a monad transformer. We use the `lift' function
-- of the monad transformer.
  
liftTP	:: (Monad (t m), Monad m, MonadTrans t) => TP m -> TP (t m)
liftTP	= msubstTP lift

liftTU	:: (Monad (t m), Monad m, MonadTrans t) => TU a m -> TU a (t m)
liftTU	= msubstTU lift

--- Remove an effect from the strategy monad --------------------------------- 

-- The first monad is obtained from the second 
-- via a monad transformer. We use the `unlift'
-- function of the monad transformer.
  
--- run the transformed monad to recover the original one
unliftTP 		:: (Monad (t m), Monad m, MonadUnTrans s t) 
			=> (forall a . s a a) -> TP (t m) -> TP m
unliftTP alg 		=  msubstTP (unlift alg)

unliftTU 		:: (Monad (t m), Monad m, MonadUnTrans s t) 
			=> s a a -> TU a (t m) -> TU a m
unliftTU alg 		=  msubstTU (unlift alg)

-- Does not work:
-- unliftS alg 		=  msubstS (unlift alg)

	
--- Localize the partiality effect -------------------------------------------

-- Safe versions provide default to cope with Nothing.
-- Note the explicit universal quantification of this value
-- in the TP case.

guaranteeSuccessTP :: (Monad (t m), Monad m, MonadUnTrans MaybeAlg t) 
                   => (forall a . a) -> TP (t m) -> TP m
guaranteeSuccessTP x s
  = unliftTP (MaybeAlg x id) s

guaranteeSuccessTU :: (Monad (t m), Monad m, MonadUnTrans MaybeAlg t) 
                   => a -> TU a (t m) -> TU a m
guaranteeSuccessTU x s
  = unliftTU (MaybeAlg x id) s



-- Does not work:
-- guaranteeSuccessS :: (Monad (t m), Monad m, MonadUnTrans MaybeAlg t) 
--                   => a -> TU a (t m) -> TU a m
-- guaranteeSuccessS x s
--   = msubstS (unlift (MaybeAlg x id)) s


-- Unsafe version uses `undefined' to cope with Nothing.
-- For the TP case, this is the only possible default value.

unsafeGuaranteeSuccessTP s
  = guaranteeSuccessTP undefined s
unsafeGuaranteeSuccessTU s
  = guaranteeSuccessTU undefined s

-- Does not work:
-- unsafeGuaranteeSuccessS s
--   = guaranteeSuccessS undefined s


--- Localize the state effect ------------------------------------------------

-- First argument represents the initial state.  

localStateTP :: (Monad (t m), Monad m, MonadUnTrans (StateAlg s) t) 
             => s -> TP (t m) -> TP m 
localStateTP s
  = unliftTP (StateAlg s fst)

localStateTU :: (Monad (t m), Monad m, MonadUnTrans (StateAlg s) t) 
             => s -> TU a (t m) -> TU a m 
localStateTU s
  = unliftTU (StateAlg s fst)

     
------------------------------------------------------------------------------
