{-----------------------------------------------------------------------------

	        A LIBRARY OF FUNCTIONAL STRATEGY COMBINATORS

		                 StrategyLib

                   Ralf Laemmel                Joost Visser
               CWI & VU, Amsterdam            CWI, Amsterdam

This module is part of a library of functional strategy combinators,
including combinators for generic traversal. This module provides
non-strategic functionality for running monads and unlifting monad
transformers.

-----------------------------------------------------------------------------}

module MonadRun where

import MonadTrans
import MonadIdentity
import MonadState
import MonadList
import MonadMaybe
import MonadError
import IOExts (unsafePerformIO) -- for running IO monads


--- Running monads -----------------------------------------------------------

-- The overloaded function run takes as first argument an `algebra' which
-- captures the ingredients necessary to run the particular monad at hand.
-- This algebra is parameterized with the domain and co-domain of run.

class MonadRun s m | m -> s where
  run :: s a b -> m a -> b

-- Identity
instance MonadRun (->) Identity where
  run alg = alg . runIdentity

-- Maybe
data MaybeAlg a b	= MaybeAlg { nothing :: b, just :: a -> b }
instance MonadRun MaybeAlg Maybe where
  run alg = maybe (nothing alg) (just alg)

-- Error
data ErrorAlg e a b 	= ErrorAlg { left :: e -> b, right :: a -> b } 
instance MonadRun (ErrorAlg e) (Either e) where
  run alg = either (left alg) (right alg)

-- List
data ListAlg a b	= ListAlg { nil :: b, cons :: a -> b -> b }
instance MonadRun ListAlg [] where
  run alg = foldr (cons alg) (nil alg)

-- State
data StateAlg s a b 	= StateAlg { --state :: (s -> (a,s)) -> b,
                                     first :: s,
				     next :: (a,s) -> b }
evalStateAlg s = StateAlg (\f -> fst (f s)) 
execStateAlg s = StateAlg (\f -> snd (f s)) 
instance MonadRun (StateAlg s) (State s) where  
  run alg = \ma -> next alg (runState ma (first alg))

-- IO
instance MonadRun (->) IO where
  run alg = alg . unsafePerformIO  

--- Exchange one monad by another --------------------------------------------
  
-- Run one monad, and put its `value' in another.
-- This is basically a monadic version of the run function itself.
-- Note that the two monads are unrelated.
mrun 		:: (MonadRun s m ,Monad m') => s a b -> m a -> m' b
mrun alg ma 	=  return (run alg ma)

--- Unlifting monad transformers ---------------------------------------------

-- Just as a base monad can be run to remove the monad, so can a transformed
-- monad be unlifted to remove the transformer and obtain the original monad.

-- The overloaded function unlift for monad transformers takes as first 
-- argument an `algebra' just like the run function for base monads. For each
-- monad transformer, the same algebra is used as for the base monad of which
-- the transformer is the parameterized variant.

class MonadUnTrans s t | t -> s where
  unlift :: Monad m => s a b -> t m a -> m b

-- List
instance MonadUnTrans ListAlg ListT where
  unlift alg ma = do as <- runListT ma 
                     return (foldr (cons alg) (nil alg) as)

-- Maybe
instance MonadUnTrans MaybeAlg MaybeT where
  unlift alg ma = do ea <- runMaybeT ma
                     return (maybe (nothing alg) (just alg) ea) 
		     
-- Error
instance MonadUnTrans (ErrorAlg e) (ErrorT e) where
  unlift alg ma = do ea <- runErrorT ma
                     return (either (left alg) (right alg) ea) 
		     	
-- State     
instance MonadUnTrans (StateAlg s) (StateT s) where
  unlift alg ma = do as <- runStateT ma (first alg)
                     return (next alg as) 

---- Auxiliary monadic functions ---------------------------------------------

-- Result of pair programming with Alberto Pardo

mchoice' :: (Monad m, MonadUnTrans MaybeAlg t) 
         => (a -> t m b) -> (a -> m b) -> a -> m b
f `mchoice'` g = \a -> do ea <- unlift (MaybeAlg Nothing Just) (f a)
			  maybe (g a) (return) ea
			  
mplus' 		:: (Monad m, MonadUnTrans MaybeAlg t) 
       		=> t m b -> m b -> m b
m1 `mplus'` m2 	=  unlift (MaybeAlg m2 return) m1 >>= id

mswitch			:: (Monad m, MonadUnTrans MaybeAlg t) 
       			=> [t m b] -> m b -> m b
mswitch [] m		=  m
mswitch (tm:tms) m	=  tm `mplus'` (mswitch tms m)		

-- in terms of foldr:
mswitch0		:: (Monad m, MonadUnTrans MaybeAlg t) 
       			=> [t m b] -> m b -> m b
mswitch0 tms m          =  foldr mplus' m tms	


-- with mplus' expanded:
mswitch1		:: (Monad m, MonadUnTrans MaybeAlg t) 
       			=> [t m b] -> m b -> m b
mswitch1 [] m		=  m
mswitch1 (tm:tms) m	=  unlift (MaybeAlg (mswitch1 tms m) return) tm >>= id 		
	
-- We could also reserve the unlift to the very end:		 
mswitch'		:: (Monad m, MonadUnTrans MaybeAlg t, 
                            MonadPlus (t m)) 
       			=> [t m b] -> m b -> m b
mswitch' tms m		=  (foldr mplus mzero tms) `mplus'` m 

-- Specialization for MaybeT:

mayswitch		:: (Monad m) => [MaybeT m b] -> m b -> m b
mayswitch tms m		=  (foldr mplus mzero tms) `mplus'` m 

mchoices fs f		=  \a -> mswitch' (map (\f -> f a) fs) (f a)


------------------------------------------------------------------------------
