{-----------------------------------------------------------------------------

	         A LIBRARY OF FUNCTIONAL STRATEGY COMBINATORS

		                 StrategyLib

                   Ralf Laemmel                Joost Visser
               CWI & VU, Amsterdam         CWI & SIG, Amsterdam

This module is part of a library of functional strategy combinators,
including combinators for generic traversal. This module defines
traversal schemes. Such schemes have formed the core of StrategyLib
since its first release. The portfolio as it stands now captures part
of the design in the paper "... Polymorphic Symphony".

-----------------------------------------------------------------------------} 



module TraversalTheme where

import StrategyPrelude
import OverloadingTheme
import FlowTheme
import Monad
import Monoid



--- Full traversals: td -- top-down; bu -- bottom-up -------------------------

full_tdTP	:: Monad m => TP m -> TP m
full_tdTP s	=  s `seqTP` (allTP (full_tdTP s))

full_buTP 	:: Monad m => TP m -> TP m
full_buTP s	=  (allTP (full_buTP s)) `seqTP` s

full_tdTU 	:: (Monad m, Monoid a) => TU a m -> TU a m
full_tdTU s	=  op2TU mappend s (allTU' (full_tdTU s))



--- Traversals with stop conditions (cut-off) --------------------------------

stop_tdTP 	:: MonadPlus m => TP m -> TP m
stop_tdTP s	=  s `choiceTP` (allTP (stop_tdTP s))

stop_tdTU 	:: (MonadPlus m, Monoid a) => TU a m -> TU a m
stop_tdTU s	=  s `choiceTU` (allTU' (stop_tdTU s))



--- Single hit traversal -----------------------------------------------------

once_tdTP 	:: MonadPlus m => TP m -> TP m
once_tdTP s	=  s `choiceTP` (oneTP (once_tdTP s))

once_tdTU 	:: MonadPlus m => TU a m -> TU a m
once_tdTU s	=  s `choiceTU` (oneTU (once_tdTU s))

once_buTP 	:: MonadPlus m => TP m -> TP m
once_buTP s	=  (oneTP (once_buTP s)) `choiceTP` s

once_buTU 	:: MonadPlus m => TU a m -> TU a m
once_buTU s	=  (oneTU (once_buTU s)) `choiceTU` s



--- Traversal with environment propagation -----------------------------------

once_peTU :: MonadPlus m => e -> (e -> TU e m) -> (e -> TU a m) -> TU a m
once_peTU e s' s = s e
                   `choiceTU`
                   (s' e `passTU` \e' -> oneTU (once_peTU e' s' s))



--- Recursive completion of one-layer traversal ------------------------------

all_recTU o s = s `o` allTU' (all_recTU o s)
one_recTU o s = s `o` oneTU (one_recTU o s)



--- Overloaded schemes for traversal -----------------------------------------
--- See the paper "... Polymorphic symphony" for a discussion ----------------

full_td 	:: StrategyMonoid s m => s m -> s m
full_td s	=  s `combS` (allS (full_td s))

once_td 	:: StrategyPlus s m => s m -> s m 
once_td s       =  s `choiceS` (oneS (once_td s))

once_bu 	:: StrategyPlus s m => s m -> s m
once_bu s       =  (oneS (once_bu s)) `choiceS` s

once_pe         :: StrategyPlus s m => (e -> s m) -> (e -> TU e m) -> e -> s m
once_pe s s' e  =  s e `choiceS` (s' e `passS` (\e' -> oneS (once_pe s s' e)))



--- One-layer traversal combinators  -----------------------------------------

{- 
   For performance and uniformity reasons, anyTP and someTP are 
   primitives, but they could have been defined as follows:
-}

anyTP'		:: MonadPlus m => TP m -> TP m
anyTP' s	=  allTP (tryTP s)

someTP'		:: MonadPlus m => TP m -> TP m
someTP' s	=  (testTP (notTP (allTP (notTP s)))) `seqTP` (anyTP s)



--- Some synonyms for convenience --------------------------------------------

topdown   :: Monad m => TP m -> TP m
topdown   =  full_tdTP

crush     :: (Monad m, Monoid u) => TU u m -> TU u m 
crush     =  full_tdTU

collect   :: Monad m => TU [a] m -> TU [a] m 
collect   =  crush

select    :: MonadPlus m => TU u m -> TU u m 
select    =  once_tdTU

selectenv :: MonadPlus m => e -> (e -> TU e m) -> (e -> TU a m) -> TU a m
selectenv =  once_peTU



------------------------------------------------------------------------------
